/**
  *  \file client/tiles.cc
  */

#include <memory>
#include "client/tiles.h"
#include "client/tiles/baseheader.h"
#include "client/tiles/basemineral.h"
#include "client/tiles/baseorder.h"
#include "client/tiles/baseoverview.h"
#include "client/tiles/basetech.h"
#include "client/tiles/errortile.h"
#include "client/tiles/nheader.h"
#include "client/tiles/objheader.h"
#include "client/tiles/planetbase.h"
#include "client/tiles/planetcolonist.h"
#include "client/tiles/planetecon.h"
#include "client/tiles/planetfcode.h"
#include "client/tiles/planetheader.h"
#include "client/tiles/planetmining.h"
#include "client/tiles/planetnatives.h"
#include "client/tiles/planetoverview.h"
#include "client/tiles/shipcargo.h"
#include "client/tiles/shipequipment.h"
#include "client/tiles/shipheader.h"
#include "client/tiles/shipmission.h"
#include "client/tiles/shipmovement.h"
#include "client/tiles/shipoverview.h"
#include "cpluslib/format.h"
#include "int/if/iterif.h"
#include "util/nls.h"



WTile*
createTile(string_t name, GObjectType& type)
{
    // Base
    if (name == "BASEHEADER")
        return new WBaseScreenHeaderTile(type);
    if (name == "BASEMINERAL")
        return new WBaseMineralTile(type);
    if (name == "BASETECH")
        return new WBaseTechTile(type);
    if (name == "BASEORDER")
        return new WBaseOrderTile(type);
    if (name == "BASECOMMENT")
        return new WPlanetBaseTile(type, true);
    if (name == "BASEOVERVIEW")
        return new WBaseOverviewTile(type);

    // Planet
    if (name == "PLANETHEADER")
        return new WPlanetScreenHeaderTile(type);
    if (name == "PLANETECONOMY")
        return new WPlanetEconomyTile(type);
    if (name == "PLANETNATIVES")
        return new WPlanetNativeTile(type);
    if (name == "PLANETCOLONISTS")
        return new WPlanetColonistTile(type);
    if (name == "PLANETFCODE")
        return new WPlanetFCodeTile(type);
    if (name == "PLANETCOMMENT")
        return new WPlanetBaseTile(type, false);
    if (name == "PLANETOVERVIEW")
        return new WPlanetOverviewTile(type);

    // Ship
    if (name == "SHIPHEADER")
        return new WShipScreenHeaderTile(type);
    if (name == "SHIPEQUIPMENT")
        return new WShipEquipmentTile(type);
    if (name == "SHIPCARGO")
        return new WShipCargoTile(type);
    if (name == "SHIPMISSION")
        return new WShipMissionTile(type);
    if (name == "SHIPMOVEMENT")
        return new WShipMovementTile(type);
    if (name == "SHIPOVERVIEW")
        return new WShipOverviewTile(type);

    // Misc
    if (name == "OBJHEADER")
        return new WObjectTypeHeaderTile(type);

    // Narrow
    if (name == "NARROWHEADER")
        return new WNarrowHeaderTile(type);
    if (name == "NARROWSHIPEQUIPMENT")
        return new WNarrowShipEquipmentTile(type);
    if (name == "NARROWSHIPCARGO")
        return new WNarrowShipCargoTile(type);
    if (name == "NARROWSHIPMISSION")
        return new WNarrowShipMissionTile(type);
    if (name == "NARROWPLANETMINERAL")
        return new WNarrowPlanetMineralTile(type);
    if (name == "NARROWPLANETECONOMY")
        return new WNarrowPlanetEconomyTile(type);
    if (name == "NARROWPLANETCOLONISTS")
        return new WNarrowPlanetColonistTile(type);
    if (name == "NARROWPLANETNATIVES")
        return new WNarrowPlanetNativeTile(type);
    if (name == "NARROWPLANETFCODE")
        return new WNarrowPlanetFCodeTile(type);
    if (name == "NARROWBASETECH")
        return new WNarrowBaseTechTile(type);
    if (name == "NARROWBASEORDER")
        return new WNarrowBaseOrderTile(type);

    return 0;
}


/***************************** Configuration *****************************/
/***************************** (preliminary) *****************************/

namespace {
    struct TileConfig {
        const char* name;       // name of tile type
        const char* title;      // may be null
        int type;
    };
}

static const TileConfig ship_screen[] = {
    { "SHIPHEADER",    0, 0 },
    { "SHIPEQUIPMENT", N_("Equipment & Crew:"), 0 },
    { "SHIPCARGO",     N_("Aboard:"), 0 },
    { "SHIPMISSION",   N_("Mission:"), 0 },
    { "SHIPMOVEMENT",  N_("Movement:"), 0 },
    { 0, 0, 0 },
};

static const TileConfig planet_screen[] = {
    { "PLANETHEADER",    0, 0 },
    { "PLANETECONOMY",   N_("Economy:"), 0 },
    { "PLANETNATIVES",   N_("Natives:"), 0 },
    { "PLANETCOLONISTS", N_("Colonists:"), 0 },
    { "PLANETFCODE",     0, 0 },
    { "PLANETCOMMENT",   0, 0 },
    { 0, 0, 0 },
};

static const TileConfig base_screen[] = {
    { "BASEHEADER",      0, 0 },
    { "BASEMINERAL",     N_("Minerals & Funds:"), 0 },
    { "BASETECH",        N_("Technology & Defense:"), 0 },
    { "BASEORDER",       N_("Orders:"), 0 },
    { "PLANETFCODE",     0, 0 },
    { "BASECOMMENT",     0, 0 },
    { 0, 0, 0 },
};

static const TileConfig ship_lock[] = {
    { "NARROWHEADER",        0, 0 },
    { "NARROWSHIPEQUIPMENT", 0, 0 },
    { "NARROWSHIPCARGO",     0, 0 },
    { "NARROWSHIPMISSION",   0, 0 },
    { 0, 0, 0 },
};

static const TileConfig planet_lock[] = {
    { "NARROWHEADER",          0, 0 },
    { "NARROWPLANETMINERAL",   0, 0 },
    { "NARROWPLANETECONOMY",   0, 0 },
    { "NARROWPLANETCOLONISTS", 0, 0 },
    { "NARROWPLANETNATIVES",   0, 0 },
    { "NARROWPLANETFCODE",     0, 0 },
    { 0, 0, 0 },
};

static const TileConfig unknown_planet_lock[] = {
    { "NARROWHEADER", 0, 0 },
    { 0, 0, 0 },
};

static const TileConfig base_lock[] = {
    { "NARROWHEADER",        0, 0 },
    { "NARROWPLANETMINERAL", 0, 0 },
    { "NARROWBASETECH",      0, 0 },
    { "NARROWBASEORDER",     0, 0 },
    { 0, 0, 0 },
};

static const TileConfig blank[] = {
    { 0, 0, 0 },
};

static const TileConfig*
getTileLayout(string_t name)
{
    if (name == "SHIPSCREEN")
        return ship_screen;
    if (name == "PLANETSCREEN")
        return planet_screen;
    if (name == "BASESCREEN")
        return base_screen;
    if (name == "PLANETLOCK")
        return planet_lock;
    if (name == "SHIPLOCK")
        return ship_lock;
    if (name == "BASELOCK")
        return base_lock;
    if (name == "UNKNOWNPLANETLOCK")
        return unknown_planet_lock;
    if (name == "")
        return blank;
    return 0;
}

void
createTileLayout(WTileContainer& container, string_t layoutName, GObjectType& baseType)
{
    const TileConfig* cfg = getTileLayout(layoutName);
    if (!cfg) {
        container.addNewTile(new WErrorTile(format(_("Error: unknown layout \"%s\""), layoutName)));
    } else {
        while (cfg->name != 0) {
            // Figure out type
            GObjectType* tileType;
            if (cfg->type == 0)
                tileType = &baseType;
            else
                tileType = getObjectTypeFromIteratorId(cfg->type);
            if (tileType == 0) {
                container.addNewTile(new WErrorTile(format(_("Error: unknown type #%d"), cfg->type)));
            } else {
                std::auto_ptr<WTile> t(createTile(cfg->name, *tileType));
                if (!t.get()) {
                    container.addNewTile(new WErrorTile(format(_("Error: unknown tile \"%s\""), cfg->name)));
                } else {
                    if (cfg->title) {
                        t->setHeading(_(cfg->title));
                        t->setDisplayState(WTile::sNormal);
                    }
                    container.addNewTile(t.release());
                }
            }
            ++cfg;
        }
    }
}
