/**
  *  \file client/tiles/baseorder.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/baseorder.h"
#include "ui/widgets.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/spec.h"
#include "game/planet.h"
#include "game/global.h"
#include "game/tables.h"
#include "game/ship.h"
#include "ui/table.h"

/***************************** WBaseOrderTile ****************************/

WBaseOrderTile::WBaseOrderTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, 100), type)
{
    setLayout(LeftJustified, 5, true);
    addControl(GfxRect(0,  2, 20, 20), h.add(new UIButton("B", 'b')));
    addControl(GfxRect(0, 27, 20, 20), h.add(new UIButton(cm_Base_Recycle, "R", 'r')).addCommand());
    addControl(GfxRect(0, 52, 20, 20), h.add(new UIButton("M", 'm')));
    addControl(GfxRect(0, 77, 20, 20), h.add(new UIButton("A", 'a')));
}

void
WBaseOrderTile::drawData(GfxCanvas& can)
{
    GPlanet* p;
    if ((p = dynamic_cast<GPlanet*>(getObject())) != 0 && p->hasBaseData()) {
        /* headers */
        setColor(UIColor::tc_Static);
        outText(can, FONT_NORMAL, 35, 55, _("Mission:"));
        outText(can, FONT_NORMAL, 35, 80, _("Build Fighters/Torpedoes"));

        /* ship */
        int hull = p->getBaseBuildHull();
        int clnid = p->getShipCloningHere(getDisplayedTurn().getCurrentUniverse());
        string_t text;

        if (hull) {
            if (clnid) {
                setColor(UIColor::tc_Red);
                text = format(_("Cloning, and building %s"),
                              getHull(hull).getName());
            } else {
                setColor(UIColor::tc_Yellow);
                text = format(_("Building a %s"),
                              getHull(hull).getName());
            }
        } else {
            if (clnid) {
                setColor(UIColor::tc_Yellow);
                text = format(_("Cloning %s"),
                              getDisplayedTurn().getCurrentUniverse().getShip(clnid).getName());
            } else {
                setColor(UIColor::tc_Static);
                text = _("Build a new ship");
            }
        }
        outTextF(can, FONT_NORMAL, 35, 5, extent.w-35, text);

        /* fix/recycle */
        if (p->getBaseShipyardOrder()) {
            setColor(UIColor::tc_Yellow);
            outTextF(can, FONT_NORMAL, 35, 30, extent.w-35,
                     format(p->getBaseShipyardOrder() == 1 /* FIXME: we have an enum for that */
                            ? _("Repairing %s")
                            : _("Recycling %s"),
                            getDisplayedTurn().getCurrentUniverse().getShip(p->getBaseShipyardId()).getFullName()));
        } else {
            setColor(UIColor::tc_Static);
            outTextF(can, FONT_NORMAL, 35, 30, extent.w-35, _("Repair/recycle a starship"));
        }

        /* mission */
        const int mx = 65 + fonts[FONT_NORMAL]->getTextWidth(_("Mission:"));
        setColor(UIColor::tc_Green);
        outTextF(can, FONT_NORMAL, mx, 55, extent.w-mx, getBaseMissionName(p->getBaseMission()));
    }
}



WNarrowBaseOrderTile::WNarrowBaseOrderTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(250, 4*font_heights[FONT_NORMAL]), type)
{ }

void
WNarrowBaseOrderTile::drawData(GfxCanvas& can)
{
    GPlanet* p = dynamic_cast<GPlanet*>(getObject());
    if (!p || !p->hasBaseData())
        return;

    /* Building foo
       Repairing foo
       Mission:  repair
       FCode:    xxx */

    UISkin& sk = getSkin();
    int x = getExtent().x;
    int y = getExtent().y;
    int w = getExtent().w;

    /* Build */
    int hull = p->getBaseBuildHull();
    int clnid = p->getShipCloningHere(getDisplayedTurn().getCurrentUniverse());
    string_t text;

    if (hull) {
        if (clnid) {
            sk.setColor(UIColor::tc_Red);
            text = format(_("Cloning, and building %s"),
                          getHull(hull).getName());
        } else {
            sk.setColor(UIColor::tc_Yellow);
            text = format(_("Building a %s"),
                          getHull(hull).getName());
        }
    } else {
        if (clnid) {
            sk.setColor(UIColor::tc_Yellow);
            text = format(_("Cloning %s"),
                          getDisplayedTurn().getCurrentUniverse().getShip(clnid).getName());
        } else {
            sk.setColor(UIColor::tc_Static);
            text = _("Not building a ship");
        }
    }
    sk.outTextF(can, FONT_NORMAL, x, y, w, text);
    y += font_heights[FONT_NORMAL];

    /* Shipyard */
    if (p->getBaseShipyardOrder()) {
        sk.setColor(UIColor::tc_Yellow);
        sk.outTextF(can, FONT_NORMAL, x, y, w, 
                    format(p->getBaseShipyardOrder() == 1 /* FIXME: we have an enum for that */
                           ? _("Repairing %s")
                           : _("Recycling %s"),
                           getDisplayedTurn().getCurrentUniverse().getShip(p->getBaseShipyardId()).getFullName()));
    } else {
        sk.setColor(UIColor::tc_Static);
        sk.outTextF(can, FONT_NORMAL, x, y, w, _("Not working on a starship"));
    }
    y += font_heights[FONT_NORMAL];

    /* Mission/FCode */
    UITableWriter tw(can, sk, GfxRect(x, y, w, getExtent().getBottomY() - y), FONT_NORMAL);
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Mission:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Friendly Code:"));

    tw.newLColumn();
    tw.writeVariableCell(UIColor::tc_Green, getBaseMissionName(p->getBaseMission()));
    tw.writeVariableCell(UIColor::tc_Green, p->getFCode());

    tw.finish();
}
