/**
  *  \file client/tiles/planetmining.cc
  */

#include "client/tiles/planetmining.h"
#include "cpluslib/format.h"
#include "game/planet.h"
#include "game/tables.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "util/misc.h"
#include "util/nls.h"
#include "ui/table.h"

WPlanetMiningTile::WPlanetMiningTile(GPlanet& planet, GCargoType ele)
    : WTile(GfxPoint(295, 4 * font_heights[FONT_NORMAL] + font_heights[FONT_TITLE] + 5)),
      planet(planet), ele(ele),
      mines(planet.getStructures(str_Mines))
{ }

static void
drawPercentBar(GfxCanvas& can, GfxRect rect, long fill, int divi, int color)
{
    int split = rect.w * fill / (100*divi);
    if (fill && !split)
        split = 1;
    if (split > rect.w)
        split = rect.w;

    drawSolidBar(can, GfxRect(rect.x, rect.y, split, rect.h), standard_colors[color]);
    drawSolidBar(can, GfxRect(rect.x+split, rect.y, rect.w-split, rect.h), standard_colors[COLOR_SHIELD+3]);
}

void
WPlanetMiningTile::drawData(GfxCanvas& can)
{
    const int wi = extent.w/2 - 5;
    int y = 3;

    drawFrameDown(can, getExtent());

    setColor(UIColor::tc_Heading);
    outTextF(can, FONT_NORMAL, 5, y, wi, getCargoName(ele));
    y += font_heights[FONT_NORMAL] + 3;

    getSkin().setSolid(true);
    setColor(UIColor::tc_Static);

    if (planet.getOreGround(ele).isKnown()
        && planet.getOreDensity(ele).isKnown()
        && planet.isCargoAmountKnown(ele))
    {
        /* Mining */
        if (mines.isKnown()) {
            int amount   = planet.getMiningCapacity(ele, mines);
            int32 ground = planet.getOreGround(ele);

            // Show mining capacity
            if (amount > ground)
                setColor(UIColor::tc_Yellow);
            else
                setColor(UIColor::tc_Green);
            getSkin().getParameters().just_x = 2;
            outTextF(can, FONT_NORMAL, extent.w-5, 3, wi, format(_("%d kt/turn"), amount));

            // Show mining time
            string_t timestr;
            if (amount != 0) {
                int32 time = (ground + (amount-1)) / amount;
                if (time > 30)
                    timestr = _(">30 turns");
                else
                    timestr = format(_("%d turn%!1{s%}"), time);
            }
            outTextF(can, FONT_NORMAL, extent.w-5, 3+font_heights[FONT_NORMAL], wi, timestr);

            getSkin().getParameters().just_x = 0;
        }

        /* Ore */
        setColor(UIColor::tc_Static);
        if (planet.getOreGround(ele))
            outTextF(can, FONT_NORMAL, 5, y, wi,
                     format("%s, %s", getMineralMassClassName(planet.getOreGround(ele)),
                            getMineralDensityClassName(planet.getOreDensity(ele))));
        else
            outTextF(can, FONT_NORMAL, 5, y, wi, _("none"));
        y += font_heights[FONT_NORMAL];

        outText(can, FONT_NORMAL, 5, y, _("mined"));
        drawPercentBar(can, GfxRect(extent.x+60, extent.y+y, 150, 15),
                       planet.getCargo(ele), 82, COLOR_WHITE);
        getSkin().getParameters().just_x = 2;
        outTextF(can, FONT_NORMAL, extent.w - 25, y, 60, numToString(planet.getCargo(ele)));
        getSkin().getParameters().just_x = 0;
        outTextF(can, FONT_NORMAL, extent.w - 25, y, 20, " kt");
        y += font_heights[FONT_NORMAL];

        outText(can, FONT_NORMAL, 5, y, _("ground"));
        drawPercentBar(can, GfxRect(extent.x+60, extent.y+y, 150, 15),
                       planet.getOreGround(ele), 82, COLOR_WHITE);
        getSkin().getParameters().just_x = 2;
        outTextF(can, FONT_NORMAL, extent.w - 25, y, 60, numToString(planet.getOreGround(ele)));
        getSkin().getParameters().just_x = 0;
        outTextF(can, FONT_NORMAL, extent.w - 25, y, 20, " kt");
        y += font_heights[FONT_NORMAL];

        outText(can, FONT_NORMAL, 5, y, _("density"));
        drawPercentBar(can, GfxRect(extent.x+60, extent.y+y, 150, 15),
                       planet.getOreDensity(ele), 1, COLOR_GRAY);
        getSkin().getParameters().just_x = 2;
        outTextF(can, FONT_NORMAL, extent.w - 25, y, 60, numToString(planet.getOreDensity(ele)));
        getSkin().getParameters().just_x = 0;
        outTextF(can, FONT_NORMAL, extent.w - 25, y, 20, " %");
        y += font_heights[FONT_NORMAL];
    } else if (planet.isCargoAmountKnown(ele)) {
        if (planet.getCargo(ele))
            outTextF(can, FONT_NORMAL, 5, y, extent.w-10, getMineralMassClassName(planet.getCargo(ele)));
        else
            outTextF(can, FONT_NORMAL, 5, y, extent.w-10, _("none"));
        y += font_heights[FONT_NORMAL];

        outText(can, FONT_NORMAL, 5, y, _("total"));
        drawPercentBar(can, GfxRect(extent.x+60, extent.y+y, 150, 15),
                       planet.getCargo(ele), 82, COLOR_WHITE);
        getSkin().getParameters().just_x = 2;
        outTextF(can, FONT_NORMAL, extent.w - 25, y, 60, numToString(planet.getCargo(ele)));
        getSkin().getParameters().just_x = 0;
        outTextF(can, FONT_NORMAL, extent.w - 25, y, 20, " kt");
        y += font_heights[FONT_NORMAL];

        outTextFormatted(can, getSkin().getParameters(), extent.x + 5, extent.y + y, extent.w - 10,
                         _("(no information on density and mined/ground masses available.)"));
    }
}

void
WPlanetMiningTile::setNumMines(mp16_t mines)
{
    this->mines = mines;
    drawWidget();
}




WNarrowPlanetMineralTile::WNarrowPlanetMineralTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(250, 5 * font_heights[FONT_NORMAL]), type)
{ }

void
WNarrowPlanetMineralTile::drawData(GfxCanvas& can)
{
    GPlanet* pl = dynamic_cast<GPlanet*>(getObject());
    if (!pl) {
        getSkin().drawBackground(can, getExtent());
        return;
    }

    enum { N = 4 };
    static const GCargoType eles[N] = { el_Neutronium, el_Tritanium, el_Duranium, el_Molybdenum };

    UITableWriter tw(can, getSkin(), getExtent(), FONT_NORMAL);
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    tw.writeFixedCell(UIColor::tc_Static, _("Neu:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Tri:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Dur:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Mol:"));

    /* Mined */
    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("9,999,999"));
    tw.writeFixedCell(UIColor::tc_Static, _("Mined"));
    for (int i = 0; i < N; ++i) {
        if (pl->isCargoAmountKnown(eles[i])) {
            tw.writeVariableCell(UIColor::tc_Green, numToString(pl->getCargoRaw(eles[i])));
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }
    tw.newLColumn(fonts[FONT_NORMAL]->getTextWidth("kt"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    for (int i = 0; i < N; ++i) {
        if (pl->isCargoAmountKnown(eles[i])) {
            tw.writeVariableCell(UIColor::tc_Green, "kt");
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }

    /* Ground */
    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("9,999,999"));
    tw.writeFixedCell(UIColor::tc_Static, _("Ground"));
    for (int i = 0; i < N; ++i) {
        if (pl->getOreGround(eles[i]).isKnown()) {
            tw.writeVariableCell(UIColor::tc_Green, numToString(pl->getOreGround(eles[i])));
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }
    tw.newLColumn(fonts[FONT_NORMAL]->getTextWidth("kt"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    for (int i = 0; i < N; ++i) {
        if (pl->getOreGround(eles[i]).isKnown()) {
            tw.writeVariableCell(UIColor::tc_Green, "kt");
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }

    /* Density */
    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("100%"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    for (int i = 0; i < N; ++i) {
        if (pl->getOreDensity(eles[i]).isKnown()) {
            tw.writeVariableCell(UIColor::tc_Green, numToString(pl->getOreDensity(eles[i])) + "%");
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }
}
