/**
  *  \file game/load-ov.cc
  *  \brief Directory Overview
  *
  *  Code to generate the directory overview, i.e. to check what's
  *  available.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/load.h"
#include "game/rst.h"
#include "game/trn.h"
#include "util/nls.h"
#include "util/console.h"
#include "util/except.h"
#include "cpluslib/format.h"
#include "io/io.h"
#include "cpluslib/misc.h"
#include "game/gen.h"

/** \class GDirectoryOverview
    \brief Overview over a directory

    This class scans a directory for usable game files.

    For each player, we can find a result file (have_RST) or an
    unpacked game directory (have_Gamedir), not both. The RST can be
    accompanied by a matching TRN (have_TRN); a turn file within a
    game directory is deliberately ignored.

    In addition, we might find a new RST along the unpack path
    (have_NewRST).

    If there is a conflict, the conflicting data is marked
    (have_Conflict); essentially, this says one can safely load all
    non-conflicting data.

    \todo have_NewRST is not yet implemented. */

/** Construct empty overview. Call scan() to fill it. */
GDirectoryOverview::GDirectoryOverview()
    : global_flags(0)
{
    std::fill_n(pp_flags, countof(pp_flags), 0);
}

/** Scan for files. This will populate this object with information
    about the specified directory.

    The basic idea is to scan for unpacked data first, and if none
    found, accept RSTs. PCC2 does not need data be unpacked before
    playing, but if there is unpacked data, we should of course use
    that instead of the RST and a probably older TRN. */
void
GDirectoryOverview::scan(Directory& dir)
{
    string_t fn;
    Ptr<Stream> s;
    GTimestamp times[NUM_PLAYERS];
    int  turns[NUM_PLAYERS];
    int  count = 0;

    this->global_flags = 0;
    
    /* do we have unpacked game data? */
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        turns[i-1] = 0;
        pp_flags[i-1] = 0;
        try {
            fn = format("gen%d.dat", i);
            s = dir.openFileNT(fn, Stream::C_READ);
            if (s) {
                TGen gen;
                if (getStructure(*s, gen) && gen.player_id == i && gen.turn > 0) {
                    turns[i-1] = gen.turn;
                    pp_flags[i-1] |= have_Gamedir;
                    times[i-1] = gen.timestamp;
                    ++count;
                } else {
                    console.write(LOG_WARN, format(_("File `%s' exists but is invalid -- ignored"), fn));
                }
            }
        }
        catch(std::exception& e) {
            console.write(LOG_WARN, fn + ": " + e.what());
        }
    }

    if (!count) {
        /* we don't have an unpacked game directory. Do we have a RST? */
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            try {
                fn = format("player%d.rst", i);
                s = dir.openFileNT(fn, Stream::C_READ);
                if (s) {
                    TResultGen gen;
                    GResultFile rst(*s);
                    s->seek(rst.getSectionOffset(GResultFile::sec_Gen));
                    if (getStructure(*s, gen)
                        && gen.player_id == i
                        && gen.turn > 0
                        && GGen(gen).isValidTimestampChecksum())
                    {
                        turns[i-1] = gen.turn;
                        times[i-1] = gen.timestamp;
                        pp_flags[i-1] |= have_RST;
                        ++count;
                    } else {
                        console.write(LOG_WARN, format(_("File `%s' exists but is invalid -- ignored"), fn));
                    }
                }
            }
            catch(std::exception& e) {
                console.write(LOG_WARN, fn + ": " + e.what());
            }
        }
    }

    if (!count)
        /* directory is empty */
        return;

    /* We have some data. Figure out which is current */
    int current_turn = 0, current_index = 0;
    for (int i = 0; i < NUM_PLAYERS; ++i) {
        if (turns[i] > current_turn) {
            current_turn = turns[i];
            current_index = i;
        }
    }
    ASSERT(current_turn);

    for (int i = 0; i < NUM_PLAYERS; ++i)
        if (turns[i] && (turns[i] != current_turn || times[i] != times[current_index]))
            pp_flags[i] |= have_Conflict;

    /* Last step: look for TRN files. */
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        try {
            fn = format("player%d.trn", i);
            s = dir.openFileNT(fn, Stream::C_READ);
            if (s) {
                GTurnfile tf(*s, false);
                const TTurnHeader& trn = tf.getTurnHeader();
                if (trn.player == i) {
                    /* when the turn is empty or stale, pretend we
                       have none at all */
                    if (trn.num_commands && times[i-1] == trn.timestamp) {
                        pp_flags[i-1] |= have_TRN;
                    }
                } else {
                    console.write(LOG_WARN, format(_("File `%s' exists but is invalid -- ignored"), fn));
                }
            }
        }
        catch(std::exception& e) {
            console.write(LOG_WARN, fn + ": " + e.what());
        }
    }

    // FIXME: should test for new RSTs in unpack path

    /* Now figure out global_flags. */
    for (int i = 0; i < NUM_PLAYERS; ++i)
        global_flags |= pp_flags[i];
}
