/**
  *  \file int/if/shipif.cc
  *  \brief Interpreter Game Interface: Ships
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/shipif.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/actions/basefix.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/mission.h"
#include "game/objtypes.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/tables.h"        // getHeadingName
#include "int/args.h"
#include "int/if/hullif.h"
#include "int/if/ifutil.h"
#include "int/if/objif.h"
#include "int/if/playerif.h"
#include "int/if/specif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "game/cspec.h"
#include "int/contextio.h"

/****************************** IntShipValue *****************************/

/** Implementation of "Ship(nn)".
    \param args function parameters
    \return value representing "Ship(nn)" */
IntValue*
IFShipGet(IntArgBlock& args)
{
    int32_t id;

    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_SHIPS))
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    return new IntShipContext(id);
}

/** Implementation of "ForEach Ship".
    \return Context representing first ship */
IntContext*
IFShipMake()
{
    if (!haveDisplayedTurn())
        return 0;

    int id = global_anyship_type.getNextNoWrap(0, false);
    if (id == 0)
        return 0;
    else
        return new IntShipContext(id);
}

/****************************** Ship Methods *****************************/

static void
IFShipSetComment(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    if (IntValue* v = args.getNext()) {
        IntStringValue sv(v->toString(false));
        int sid = obj.getId();
        getShipProperties(sid).set(sp_Comment, &sv);
        global_anyship_type.getObject(sid).markDirty();
    }
}

static void
IFShipSetFCode(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), ispFCode, args.getNext());
}

static void
IFShipSetEnemy(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), ispEnemyId, args.getNext());
}

static void
IFShipSetSpeed(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), ispSpeedId, args.getNext());
}

static void
IFShipSetName(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), ispName, args.getNext());
}

static void
IFShipSetMission(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int32_t m;
    int32_t i = 0, t = 0;

    // Evaluate arguments
    args.checkArgumentCount(1, 3);
    if (!checkIntArg(m, args.getNext(), 0, 10000))
        return;
    checkIntArg(i, args.getNext(), 0, 10000);
    checkIntArg(t, args.getNext(), 0, 10000);

    // Set mission on ship
    GShip& sh = static_cast<GShip&>(obj);
    if (sh.getKind() == GShip::CurrentShip) {
        sh.setMission(m, i, t);
    } else {
        throw IntError::notAssignable();
    }
}

static void
setBaseOrder(TShipyardAction order, GShip& sh)
{
    // Find the base we're at and verify the command
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    int pid = univ.getPlanetAt(sh.getPos());
    if (GError e = checkBaseAction(order, univ, sh.getId(), pid))
        throw e;

    // Do it
    univ.getPlanet(pid).setBaseShipyardOrder(order, sh.getId());
    sh.markDirty();
}

static void
IFShipFixShip(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    setBaseOrder(sya_Fix, static_cast<GShip&>(obj));
}

static void
IFShipRecycleShip(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    setBaseOrder(sya_Recycle, static_cast<GShip&>(obj));
}

static void
IFShipSetWaypoint(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int32_t x, y;

    // Evaluate arguments
    args.checkArgumentCount(2);
    if (!checkIntArg(x, args.getNext(), 0, 10000))
        return;
    if (!checkIntArg(y, args.getNext(), 0, 10000))
        return;

    // Set ship waypoint
    GShip& sh = static_cast<GShip&>(obj);
    if (sh.getKind() == GShip::CurrentShip) {
        sh.setWaypoint(GPoint(x, y).getNearestTo(sh.getPos()));
        if (sh.getMission() == GMission::msn_Intercept)
            sh.setMission(0, 0, 0);
    } else {
        throw IntError::notAssignable();
    }
}

/***************************** IntShipContext ****************************/

namespace {
    enum { ShipProperty, HullProperty, ComponentProperty, OwnerProperty, EnemyProperty, ShipMethod };
}

static const TIntMapping ship_mapping[] = {
    { "AUX",                       ispAuxName,                   ShipProperty },
    { "AUX$",                      ispAuxId,                     ShipProperty },
    { "AUX.AMMO",                  ispAuxAmmo,                   ShipProperty },
    { "AUX.COUNT",                 ispAuxCount,                  ShipProperty },
    { "AUX.SHORT",                 ispAuxShort,                  ShipProperty },
    { "BEAM",                      ispBeamName,                  ShipProperty },
    { "BEAM$",                     ispBeamId,                    ShipProperty },
    { "BEAM.COUNT",                ispBeamCount,                 ShipProperty },
    { "BEAM.MAX",                  ihpMaxBeams,                  HullProperty },
    { "BEAM.SHORT",                ispBeamShort,                 ShipProperty },
    { "CARGO.COLONISTS",           ispCargoColonists,            ShipProperty },
    { "CARGO.D",                   ispCargoD,                    ShipProperty },
    { "CARGO.FREE",                ispCargoFree,                 ShipProperty },
    { "CARGO.M",                   ispCargoM,                    ShipProperty },
    { "CARGO.MAX",                 ihpMaxCargo,                  HullProperty },
    { "CARGO.MAXFUEL",             ihpMaxFuel,                   HullProperty },
    { "CARGO.MONEY",               ispCargoMoney,                ShipProperty },
    { "CARGO.N",                   ispCargoN,                    ShipProperty },
    { "CARGO.STR",                 ispCargoStr,                  ShipProperty },
    { "CARGO.SUPPLIES",            ispCargoSupplies,             ShipProperty },
    { "CARGO.T",                   ispCargoT,                    ShipProperty },
    { "CREW",                      ispCrew,                      ShipProperty },
    { "CREW.NORMAL",               ihpMaxCrew,                   HullProperty },
    { "DAMAGE",                    ispDamage,                    ShipProperty },
    { "ENEMY",                     iplShortName,                 EnemyProperty },
    { "ENEMY$",                    ispEnemyId,                   ShipProperty },
    { "ENEMY.ADJ",                 iplAdjName,                   EnemyProperty },
    { "ENGINE",                    ispEngineName,                ShipProperty },
    { "ENGINE$",                   ispEngineId,                  ShipProperty },
    { "ENGINE.COUNT",              ihpNumEngines,                HullProperty },
    { "FCODE",                     ispFCode,                     ShipProperty },
    { "FIGHTER.BAYS",              ispFighterBays,               ShipProperty },
    { "FIGHTER.COUNT",             ispFighterCount,              ShipProperty },
    { "FIXSHIP",                   8,                            ShipMethod },
    { "FLEET",                     ispFleet,                     ShipProperty },
    { "FLEET$",                    ispFleetId,                   ShipProperty },
    { "FLEET.NAME",                ispFleetName,                 ShipProperty },
    { "FLEET.STATUS",              ispFleetStatus,               ShipProperty },
    { "HEADING",                   ispHeadingName,               ShipProperty },
    { "HEADING$",                  ispHeadingAngle,              ShipProperty },
    { "HULL",                      icpName,                      ComponentProperty },
    { "HULL$",                     icpId,                        ComponentProperty },
    { "HULL.SHORT",                icpNameShort,                 ComponentProperty },
    { "HULL.SPECIAL",              ispHullSpecial,               ShipProperty },
    { "ID",                        ispId,                        ShipProperty },
    { "LEVEL",                     ispLevel,                     ShipProperty },
    { "LOC",                       ispLoc,                       ShipProperty },
    { "LOC.X",                     ispLocX,                      ShipProperty },
    { "LOC.Y",                     ispLocY,                      ShipProperty },
    { "MARK",                      0,                            ShipMethod },
    { "MARKED",                    ispMarked,                    ShipProperty },
    { "MASS",                      ispMass,                      ShipProperty },
    { "MISSION",                   ispMissionName,               ShipProperty },
    { "MISSION$",                  ispMissionId,                 ShipProperty },
    { "MISSION.INTERCEPT",         ispMissionIntercept,          ShipProperty },
    { "MISSION.SHORT",             ispMissionShort,              ShipProperty },
    { "MISSION.TOW",               ispMissionTow,                ShipProperty },
    { "MOVE.ETA",                  ispMoveETA,                   ShipProperty },
    { "MOVE.FUEL",                 ispMoveFuel,                  ShipProperty },
    { "NAME",                      ispName,                      ShipProperty },
    { "ORBIT",                     ispOrbitName,                 ShipProperty },
    { "ORBIT$",                    ispOrbitId,                   ShipProperty },
    { "OWNER",                     iplShortName,                 OwnerProperty },
    { "OWNER$",                    iplId,                        OwnerProperty },
    { "OWNER.ADJ",                 iplAdjName,                   OwnerProperty },
    { "OWNER.REAL",                ispRealOwner,                 ShipProperty },
    { "PLAYED",                    ispPlayed,                    ShipProperty },
    { "RECYCLESHIP",               9,                            ShipMethod },
    { "SETCOMMENT",                2,                            ShipMethod },
    { "SETENEMY",                  4,                            ShipMethod },
    { "SETFCODE",                  3,                            ShipMethod },
    { "SETMISSION",                7,                            ShipMethod },
    { "SETNAME",                   6,                            ShipMethod },
    { "SETSPEED",                  5,                            ShipMethod },
    { "SETWAYPOINT",               10,                           ShipMethod },
    { "SPEED",                     ispSpeedName,                 ShipProperty },
    { "SPEED$",                    ispSpeedId,                   ShipProperty },
    { "TASK",                      ispTask,                      ShipProperty },
    { "TECH.HULL",                 icpTech,                      ComponentProperty },
    { "TORP",                      ispTorpName,                  ShipProperty },
    { "TORP$",                     ispTorpId,                    ShipProperty },
    { "TORP.COUNT",                ispTorpCount,                 ShipProperty },
    { "TORP.LCOUNT",               ispTorpLCount,                ShipProperty },
    { "TORP.LMAX",                 ihpMaxTorpLaunchers,          HullProperty },
    { "TORP.SHORT",                ispTorpShort,                 ShipProperty },
    { "TRANSFER.SHIP",             ispTransferShip,              ShipProperty },
    { "TRANSFER.SHIP.COLONISTS",   ispTransferShipColonists,     ShipProperty },
    { "TRANSFER.SHIP.D",           ispTransferShipD,             ShipProperty },
    { "TRANSFER.SHIP.ID",          ispTransferShipId,            ShipProperty },
    { "TRANSFER.SHIP.M",           ispTransferShipM,             ShipProperty },
    { "TRANSFER.SHIP.N",           ispTransferShipN,             ShipProperty },
    { "TRANSFER.SHIP.NAME",        ispTransferShipName,          ShipProperty },
    { "TRANSFER.SHIP.SUPPLIES",    ispTransferShipSupplies,      ShipProperty },
    { "TRANSFER.SHIP.T",           ispTransferShipT,             ShipProperty },
    { "TRANSFER.UNLOAD",           ispTransferUnload,            ShipProperty },
    { "TRANSFER.UNLOAD.COLONISTS", ispTransferUnloadColonists,   ShipProperty },
    { "TRANSFER.UNLOAD.D",         ispTransferUnloadD,           ShipProperty },
    { "TRANSFER.UNLOAD.ID",        ispTransferUnloadId,          ShipProperty },
    { "TRANSFER.UNLOAD.M",         ispTransferUnloadM,           ShipProperty },
    { "TRANSFER.UNLOAD.N",         ispTransferUnloadN,           ShipProperty },
    { "TRANSFER.UNLOAD.NAME",      ispTransferUnloadName,        ShipProperty },
    { "TRANSFER.UNLOAD.SUPPLIES",  ispTransferUnloadSupplies,    ShipProperty },
    { "TRANSFER.UNLOAD.T",         ispTransferUnloadT,           ShipProperty },
    { "TYPE",                      ispTypeStr,                   ShipProperty },
    { "TYPE.SHORT",                ispTypeChar,                  ShipProperty },
    { "UNMARK",                    1,                            ShipMethod },
    { "WAYPOINT",                  ispWaypointName,              ShipProperty },
    { "WAYPOINT.DIST",             ispWaypointDistance,          ShipProperty },
    { "WAYPOINT.DX",               ispWaypointDX,                ShipProperty },
    { "WAYPOINT.DY",               ispWaypointDY,                ShipProperty },
    { "WAYPOINT.PLANET",           ispWaypointPlanetId,          ShipProperty },
    { "WAYPOINT.X",                ispWaypointX,                 ShipProperty },
    { "WAYPOINT.Y",                ispWaypointY,                 ShipProperty },
};

static const IntObjectProcedureValue::call_t ship_methods[] = {
    IFObjMark,                  // 0
    IFObjUnmark,                // 1
    IFShipSetComment,           // 2
    IFShipSetFCode,             // 3
    IFShipSetEnemy,             // 4
    IFShipSetSpeed,             // 5
    IFShipSetName,              // 6
    IFShipSetMission,           // 7
    IFShipFixShip,              // 8
    IFShipRecycleShip,          // 9
    IFShipSetWaypoint,          // 10
//  (id: CMD_CARGOTRANSFER; argspec:$23;      exec:@Ship_CargoTransfer),
//  (id: CMD_CARGOUNLOAD; argspec:$12;        exec:@Ship_CargoUnload),
//  (id: CMD_CARGOUPLOAD; argspec:$12;        exec:@Ship_CargoUpload),
//  (id: CMD_SETFLEET;    argspec:ci_IntFunc; exec:@Ship_SetFleet),
};

IntShipContext::IntShipContext(int sid)
    : sid(sid)
{ }

IntShipContext::~IntShipContext()
{ }

int32_t
IntShipContext::lookup(string_t name)
{
    static const int32_t nproperties = countof(ship_mapping);

    /* Check prefix */
    if (name.compare(0, 5, "SHIP.", 5) == 0)
        name.erase(0, 5);

    /* Check user-defined properties */
    IntVariableNames::index_t ix = ship_property_names.lookup(name);
    if (ix != ship_property_names.nil)
        return ix + nproperties;

    /* Check predefined properties */
    return lookupName(name, ship_mapping, nproperties);
}

void
IntShipContext::set(int32_t index, IntValue* value)
{
    if (index < int32_t(countof(ship_mapping))) {
        /* Builtin property */
        GShip& sh = global_anyship_type.getObject(sid);
        switch (ship_mapping[index].domain) {
         case ShipProperty:
            setShipProperty(sh, IntShipProperty(ship_mapping[index].index), value);
            break;
         default:
            throw IntError::notAssignable();
        }
    } else {
        /* User property */
        getShipProperties(sid).set(index - countof(ship_mapping), value);
        global_anyship_type.getObject(sid).markDirty();
    }
}

IntValue*
IntShipContext::get(int32_t index)
{
    if (index < int32_t(countof(ship_mapping))) {
        /* Builtin property */
        GShip& sh = global_anyship_type.getObject(sid);
        switch (ship_mapping[index].domain) {
         case ShipProperty:
            return getShipProperty(sh, IntShipProperty(ship_mapping[index].index));
         case HullProperty:
            if (sh.getHullId() > 0)
                return getHullProperty(sh.getHull(), IntHullProperty(ship_mapping[index].index));
            else
                return 0;
         case ComponentProperty:
            if (sh.getHullId() > 0)
                return getComponentProperty(sh.getHull(), IntComponentProperty(ship_mapping[index].index));
            else
                return 0;
         case OwnerProperty:
            return getPlayerProperty(sh.getOwner(), IntPlayerProperty(ship_mapping[index].index));
         case EnemyProperty:
            if (sh.getPE().isKnown())
                return getPlayerProperty(sh.getPE(), IntPlayerProperty(ship_mapping[index].index));
            else
                return 0;
         case ShipMethod:
            return new IntObjectProcedureValue(sh, ship_methods[ship_mapping[index].index]);
         default:
            return 0;
        }
    } else {
        /* User property */
        return IntValue::cloneOf(getShipProperties(sid)[index - countof(ship_mapping)]);
    }
}

bool
IntShipContext::next()
{
    int next = global_anyship_type.getNextNoWrap(sid, false);
    if (next != 0) {
        sid = next;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntShipContext::clone()
{
    return new IntShipContext(sid);
}

GObject*
IntShipContext::getObject()
{
    return &global_anyship_type.getObject(sid);
}

string_t
IntShipContext::toString(bool /*readable*/)
{
    return format("Ship(%d)", sid);
}

void
IntShipContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Ship;
    sv.value = sid;
}

/** Classify ship. This yields the ship's category as a string, 0 if unknown. */
static const char*
classifyShip(const GShip& sh)
{
    int beams = sh.getBeamCount().isKnown()         ? sh.getBeamCount()         : sh.getHullId() > 0 ? sh.getHull().getMaxBeams()     : -1;
    int tubes = sh.getTorpLauncherCount().isKnown() ? sh.getTorpLauncherCount() : sh.getHullId() > 0 ? sh.getHull().getMaxLaunchers() : -1;
    int bays  = sh.getBayCount().isKnown()          ? sh.getBayCount()          : sh.getHullId() > 0 ? sh.getHull().getNumBays()      : -1;

    if (bays > 0)
        return "Carrier";
    else if (tubes > 0)
        return "Torpedo Ship";
    else if (beams > 0)
        return "Beam Weapons";
    else if (tubes == 0 && beams == 0 && bays == 0)
        return "Freighter";
    else
        return 0;
}

static IntValue*
makeIntFromMaybe(mp16_t value)
{
    if (value.isKnown())
        return makeIntValue(value);
    else
        return 0;
}

/** Get ship property.
    \param sh Ship to query
    \param isp Property to query */
IntValue*
getShipProperty(const GShip& sh, const IntShipProperty isp)
{
    switch (isp) {
     case ispAuxId:
        /* 11 for fighters, torpedo tech if any, otherwise null */
        if (sh.getBayCount().isKnown() && sh.getBayCount() > 0)
            return makeIntValue(11);
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeIntValue(sh.getTorpType());
        else
            return 0;
     case ispAuxAmmo:
        /* ammo */
        if ((sh.getBayCount().isKnown() && sh.getBayCount() > 0) || (sh.getTorpType().isKnown() && sh.getTorpType() > 0))
            return makeIntValue(sh.getAmmo());
        else
            return 0;
     case ispAuxCount:
        /* number of aux weapons */
        if (sh.getBayCount().isKnown() && sh.getBayCount() > 0)
            return makeIntValue(sh.getBayCount());
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0 && sh.getTorpLauncherCount().isKnown())
            return makeIntValue(sh.getTorpLauncherCount());
        else
            return 0;
     case ispAuxShort:
        /* short name of aux weapon */
        if (sh.getBayCount().isKnown() && sh.getBayCount() > 0)
            return makeStringValue("Ftr");
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getShortName());
        else
            return 0;
     case ispAuxName:
        /* full name of aux weapon */
        if (sh.getBayCount().isKnown() && sh.getBayCount() > 0)
            return makeStringValue("Fighters");
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getName());
        else
            return 0;
     case ispBeamId:
        /* beam type number */
        return makeIntFromMaybe(sh.getBeamType());
     case ispBeamCount:
        /* beam count */
        return makeIntFromMaybe(sh.getBeamCount());
     case ispBeamShort:
        /* short name of beam */
        if (sh.getBeamType().isKnown() && sh.getBeamType() > 0)
            return makeStringValue(::getBeam(sh.getBeamType()).getShortName());
        else
            return 0;
     case ispBeamName:
        /* beam type */
        if (sh.getBeamType().isKnown() && sh.getBeamType() > 0)
            return makeStringValue(::getBeam(sh.getBeamType()).getName());
        else
            return 0;
     case ispCargoColonists:
        /* colonists */
        if (sh.getCargoRaw(el_Colonists) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Colonists));
        else
            return 0;
     case ispCargoD:
        /* duranium */
        if (sh.getCargoRaw(el_Duranium) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Duranium));
        else
            return 0;
     case ispCargoFree:
        /* free cargo */
        return makeIntValue(sh.getFreeCargo());
     case ispCargoM:
        /* molybdenum */
        if (sh.getCargoRaw(el_Molybdenum) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Molybdenum));
        else
            return 0;
     case ispCargoMoney:
        /* money */
        if (sh.getCargoRaw(el_Money) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Money));
        else
            return 0;
     case ispCargoN:
        /* fuel */
        if (sh.getCargoRaw(el_Neutronium) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Neutronium));
        else
            return 0;
     case ispCargoStr:
        /* cargo string */
        {
            GCargoSpec sp;
            bool hadAny = false;
            for (GCargoType i = el_Neutronium; i <= el_Money; ++i) {
                if (sh.getCargoRaw(i) >= 0) {
                    sp.set(i, sh.getCargoRaw(i));
                    hadAny = true;
                }
            }
            if (sh.getTorpType().isKnown() && sh.getTorpType() > 0) {
                sp.set(el_Torps, sh.getAmmo());
            }

            if (hadAny) {
                return makeStringValue(sp.toCargoSpecString());
            } else {
                return 0;
            }
        }
     case ispCargoSupplies:
        /* supplies */
        if (sh.getCargoRaw(el_Supplies) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Supplies));
        else
            return 0;
     case ispCargoT:
        /* supplies */
        if (sh.getCargoRaw(el_Tritanium) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Tritanium));
        else
            return 0;
     case ispCrew:
        /* crew */
        return makeIntFromMaybe(sh.getCrew());
     case ispDamage:
        /* damage */
        return makeIntFromMaybe(sh.getDamage());
     case ispEnemyId:
        /* enemy */
        return makeIntFromMaybe(sh.getPE());
     case ispEngineId:
        /* engine id */
        return makeIntFromMaybe(sh.getEngine());
     case ispEngineName:
        /* engine name */
        if (sh.getEngine().isKnown())
            return makeStringValue(::getEngine(sh.getEngine()).getName());
        else
            return 0;
     case ispFCode:
        /* fcode */
        // FIXME:
        if (sh.getFCode().size())
            return makeStringValue(sh.getFCode());
        else
            return 0;
     case ispFighterBays:
        /* fighter bays */
        return makeIntFromMaybe(sh.getBayCount());
     case ispFighterCount:
        /* fighter count */
        if (sh.getBayCount().isKnown() && sh.getBayCount() > 0 && sh.getAmmo() >= 0)
            return makeIntValue(sh.getAmmo());
        else
            return 0;
     case ispFleetId:
     case ispFleetName:
     case ispFleetStatus:
     case ispFleet:
        /* Fleet stuff FIXME */
        return 0;
     case ispHeadingAngle:
        /* heading angle */
        return makeIntFromMaybe(sh.getHeading());
     case ispHeadingName:
        /* heading name */
        // FIXME: this goes through NLS. Should it?
        if (sh.getHeading().isKnown())
            return makeStringValue(getHeadingName(sh.getHeading()));
        else
            return 0;
     case ispHullSpecial:
        /* Special flags */
        {
            string_t result;
            if (sh.isExistant()) {
                if (sh.canDoSpecial(hf_Cloak) || sh.canDoSpecial(hf_AdvancedCloak) || sh.canDoSpecial(hf_HardenedCloak))
                    result += "C";
                if (sh.canDoSpecial(hf_Hyperdrive))
                    result += "H";
                if (sh.canDoSpecial(hf_Gravitonic))
                    result += "G";
                if (sh.canDoSpecial(hf_Bioscan) || sh.canDoSpecial(hf_FullBioscan))
                    result += "B";
                if (sh.canDoSpecial(hf_MerlinAlchemy) || sh.canDoSpecial(hf_AriesRefinery) || sh.canDoSpecial(hf_NeutronicRefinery))
                    result += "A";
            }
            return makeStringValue(result);
        }
     case ispId:
        /* ship Id */
        return makeIntValue(sh.getId());
     case ispLevel:
        {
            int16_t level = sh.unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel));
            if (level >= 0)
                return makeIntValue(level);
            else
                return 0;
        }
     case ispLocX:
        /* ship location */
        if (sh.isExistant())
            return makeIntValue(sh.getPos().x);
        else
            return 0;
     case ispLocY:
        /* ship location */
        if (sh.isExistant())
            return makeIntValue(sh.getPos().y);
        else
            return 0;
     case ispLoc:
        if (sh.isExistant())
            return makeStringValue(getDisplayedTurn().getCurrentUniverse().getLocationName(sh.getPos(), 0));
        else
            return 0;
     case ispMarked:
        /* marked */
        return makeBoolValue(sh.isMarked());
     case ispMass:
        /* mass */
        if (sh.isExistant())
            return makeIntValue(sh.getMass());
        else
            return 0;
     case ispMissionId:
        /* mission */
        return makeIntFromMaybe(sh.getMission());
     case ispMissionIntercept:
        /* Intercept Id */
        return makeIntFromMaybe(sh.getIntId());
     case ispMissionShort:
        /* Short mission name */
        if (sh.getMission().isKnown()) {
            const GMission* m = GMissionList::getInstance().getMissionByNumber(sh.getMission(), GPlayerSet(sh.getRealOwner()));
            if (m != 0) {
                return makeStringValue(m->getShortName());
            } else {
                return makeStringValue(format("MIT %d", int(sh.getMission())));
            }
        } else {
            return 0;
        }
     case ispMissionTow:
        /* Tow Id */
        return makeIntFromMaybe(sh.getTowId());
     case ispMissionName:
        /* mission name */
        if (sh.getMission().isKnown()) {
            const GMission* m = GMissionList::getInstance().getMissionByNumber(sh.getMission(), GPlayerSet(sh.getRealOwner()));
            if (m != 0) {
                return makeStringValue(m->getName());
            } else if (sh.getIntId().isKnown() && sh.getTowId().isKnown()) {
                return makeStringValue(format("M.I.T. %d (%d,%d)",
                                              int(sh.getMission()),
                                              int(sh.getIntId()),
                                              int(sh.getTowId())));
            } else {
                return makeStringValue(format("M.I.T. %d", int(sh.getMission())));
            }
        } else {
            return 0;
        }
     case ispMoveETA:
        if (sh.getKind() == sh.CurrentShip) {
            GShipTurnPredictor pred(getDisplayedTurn().getCurrentUniverse(), sh.getId());
            pred.computeMovement();
            return makeIntValue(pred.getTurnCount());
        } else {
            return 0;
        }
     case ispMoveFuel:
        if (sh.getKind() == sh.CurrentShip) {
            GShipTurnPredictor pred(getDisplayedTurn().getCurrentUniverse(), sh.getId());
            pred.computeMovement();
            return makeIntValue(pred.getFuelUsed());
        } else {
            return 0;
        }
     case ispName:
        if (sh.isExistant())
            return makeStringValue(sh.getName());
        else
            return 0;
     case ispOrbitId:
        if (sh.isExistant())
            return makeIntValue(sh.getOrbit());
        else
            return 0;
     case ispOrbitName:
        if (sh.isExistant() && sh.getOrbit() > 0)
            return makeStringValue(getDisplayedTurn().getCurrentUniverse().getPlanet(sh.getOrbit()).getName());
        else
            return 0;
     case ispPlayed:
        return makeBoolValue(sh.isPlayed());
     case ispRealOwner:
        return makeIntValue(sh.getRealOwner());
     case ispSpeedId:
        return makeIntFromMaybe(sh.getWarp());
     case ispSpeedName:
        if (sh.getWarp().isKnown()) {
            if (sh.isHyperdriving())
                return makeStringValue("Hyperdrive");
            else
                return makeStringValue(format("Warp %d", sh.getWarp()));
        } else {
            return 0;
        }
     case ispTask:
        // FIXME:
        return 0;
     case ispTorpId:
        return makeIntFromMaybe(sh.getTorpType());
     case ispTorpCount:
        if (sh.getTorpType().isKnown()) {
            if (sh.getTorpType() > 0)
                return makeIntValue(sh.getAmmo());
            else
                return makeIntValue(0);
        } else {
            return 0;
        }
     case ispTorpLCount:
        return makeIntFromMaybe(sh.getTorpLauncherCount());
     case ispTorpShort:
        if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getShortName());
        else
            return 0;
     case ispTorpName:
        if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getName());
        else
            return 0;
     case ispTransferShipColonists:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).colonists);
        else
            return 0;
     case ispTransferShipD:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Duranium]);
        else
            return 0;
     case ispTransferShipId:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).id);
        else
            return 0;
     case ispTransferShipM:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Molybdenum]);
        else
            return 0;
     case ispTransferShipN:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Neutronium]);
        else
            return 0;
     case ispTransferShipName:
        if (sh.getKind() == sh.CurrentShip
            && sh.isTransporterActive(sh.t_Transfer)
            && getDisplayedTurn().getCurrentUniverse().isValidShipId(sh.getTransporter(sh.t_Transfer).id))
        {
            return makeStringValue(getDisplayedTurn().getCurrentUniverse().getShip(sh.getTransporter(sh.t_Transfer).id).getName());
        } else {
            return 0;
        }
     case ispTransferShipT:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Tritanium]);
        else
            return 0;
     case ispTransferShipSupplies:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).supplies);
        else
            return 0;
     case ispTransferShip:
        if (sh.getKind() == sh.CurrentShip)
            return makeBoolValue(sh.isTransporterActive(sh.t_Transfer));
        else
            return 0;
     case ispTransferUnloadColonists:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).colonists);
        else
            return 0;
     case ispTransferUnloadD:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Duranium]);
        else
            return 0;
     case ispTransferUnloadId:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).id);
        else
            return 0;
     case ispTransferUnloadM:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Molybdenum]);
        else
            return 0;
     case ispTransferUnloadN:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Neutronium]);
        else
            return 0;
     case ispTransferUnloadName:
        if (sh.getKind() == sh.CurrentShip && sh.isTransporterActive(sh.t_Unload)) {
            if (sh.getTransporter(sh.t_Unload).id == 0)
                return makeStringValue("Jettison");
            else if (getDisplayedTurn().getCurrentUniverse().isValidPlanetId(sh.getTransporter(sh.t_Unload).id))
                return makeStringValue(getDisplayedTurn().getCurrentUniverse().getPlanet(sh.getTransporter(sh.t_Unload).id).getName());
            else
                return 0;
        } else {
            return 0;
        }
     case ispTransferUnloadT:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Tritanium]);
        else
            return 0;
     case ispTransferUnloadSupplies:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).supplies);
        else
            return 0;
     case ispTransferUnload:
        if (sh.getKind() == sh.CurrentShip)
            return makeBoolValue(sh.isTransporterActive(sh.t_Unload));
        else
            return 0;
     case ispTypeChar:
        if (const char* p = classifyShip(sh))
            return makeStringValue(string_t(p, 1));
        else
            return 0;
     case ispTypeStr:
        if (const char* p = classifyShip(sh))
            return makeStringValue(p);
        else
            return 0;
     case ispWaypointDistance:
        if (sh.getKind() == sh.CurrentShip)
            return makeFloatValue(std::sqrt(double(square(sh.getWaypointDX()) + square(sh.getWaypointDY()))));
        else
            return 0;
     case ispWaypointDX:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getWaypointDX());
        else
            return 0;
     case ispWaypointDY:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getWaypointDY());
        else
            return 0;
     case ispWaypointPlanetId:
        if (sh.getKind() == sh.CurrentShip) {
            return makeIntValue(getDisplayedTurn().getCurrentUniverse().getPlanetAt(sh.getWaypoint(), false));
        } else {
            return 0;
        }
     case ispWaypointX:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getWaypointDX() + sh.getPos().x);
        else
            return 0;
     case ispWaypointY:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getWaypointDY() + sh.getPos().y);
        else
            return 0;
     case ispWaypointName:
        if (sh.getKind() == sh.CurrentShip) {
            // FIXME: PCC 1.x also handles Intercept here
            if (sh.getWaypointDX() == 0 && sh.getWaypointDY() == 0)
                return makeStringValue("(Location)");
            else
                return makeStringValue(getDisplayedTurn().getCurrentUniverse().getLocationName(sh.getWaypoint().getNormalized(), 0));
        } else {
            return 0;
        }
    }
    return 0;
}

/** Modify ship property.
    \param sh Ship to modify
    \param isp Property to modify
    \param value New value */
void
setShipProperty(GShip& sh, const IntShipProperty isp, IntValue* value)
{
    string_t sv;
    int32_t iv;

    // Everything is only assignable for own ships. As an exception, the name is also assignable for targets.
    if (sh.getKind() != GShip::CurrentShip) {
        if (isp == ispName && sh.getKind() != GShip::NoShip) {
            // accept
        } else {
            throw IntError::notAssignable();
        }
    }

    switch (isp) {
     case ispFCode:
        if (checkStringArg(sv, value))
            sh.setFCode(sv);
        break;
     case ispMissionId:
     case ispMissionIntercept:
     case ispMissionTow:
        if (checkIntArg(iv, value, 0, 10000)) {
            int m = (isp == ispMissionId        ? iv : sh.getMission());
            int i = (isp == ispMissionIntercept ? iv : sh.getIntId());
            int t = (isp == ispMissionTow       ? iv : sh.getTowId());
            sh.setMission(m, i, t);
        }
        break;
     case ispName:
        if (checkStringArg(sv, value))
            sh.setName(sv);
        break;
     case ispSpeedId:
        if (checkIntArg(iv, value, 0, 9))
            sh.setWarp(iv);
        break;
     case ispEnemyId:
        if (checkIntArg(iv, value, 0, NUM_PLAYERS))
            sh.setPE(iv);
        break;
     default:
        throw IntError::notAssignable();
    }
}
