/**
  *  \file tests/testexpr.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <iostream>
#include "int/parseexpr.h"
#include "cpluslib/string.h"
#include "int/lex.h"
#include "cpluslib/format.h"
#include "int/exec.h"
#include "int/intglobal.h"
#include "int/if/initif.h"
#include "util/console.h"
#include "io/dirs.h"
#include "game/global.h"
#include "game/config.h"
#include "game/spec.h"
#include "game/fcode.h"
#include "game/mission.h"
#include "game/racename.h"
#include "game/coord.h"
#include "client/chartusr.h"
#include "game/team.h"
#include "game/load.h"
#include "game/stat.h"
#include "int/interror.h"
#include "game/reg.h"
#include "int/if/conif.h"
#include "game/gamedirs.h"
#include "int/optimize.h"
#include "int/process.h"

static const ModuleInitializer*const testexpr_modules[] = {
    mod_base_console,
    mod_int,
    mod_intgame_common,
    mod_intgame_console,
    mod_racenames,
    mod_spec,
    mod_team,
    mod_missions,
    mod_reg,
    mod_stat,
    mod_fcode,
    mod_config,
    0
};

static void
disassemble(IntBytecodeObject& bco)
{
    int lastLine = 0;
    for (IntBytecodeObject::pc_t i = 0; i < bco.getNumInstructions(); ++i) {
        int lineNo = bco.getLineNumber(i);
        if (lineNo != 0 && lineNo != lastLine) {
            std::cout << (format("%4d: <%02X.%02X.%04X>     %-50s ; line %d\n", i, bco(i).major, bco(i).minor, bco(i).arg)
                          << bco.getDisassembly(i)
                          << lineNo);
            lastLine = lineNo;
        } else {
            std::cout << format("%4d: <%02X.%02X.%04X>     ", i, bco(i).major, bco(i).minor, bco(i).arg)
                      << bco.getDisassembly(i) << "\n";
        }
    }
}


int main(int argc, char** argv)
{
    (void) argc;
    (void) argv;

    performInitialisation(OnStartup, testexpr_modules);

    string_t line;

#if 1
    if (argv[1]) {
        try {
            initGameGlobals(Directory::getDirName(argv[1]), Platform::getDataDirectory());
            performInitialisation(OnEnterGame, testexpr_modules);
            initCoordinates();

            Ptr<Stream> pxy = spec_file_dir->openFile("xyplan.dat", Stream::C_READ);
            getCurrentTurn().init();
            getCurrentTurn().loadPlanetXY(*pxy);

            // HACK!
            int pid = atoi(Directory::getFileName(argv[1]).c_str() + 6);

            Ptr<Stream> rst = Stream::openFile(argv[1], Stream::C_READ);
            loadResult(getCurrentTurn(), *rst, pid);

            setPlayerIds(pid, pid);
            getCurrentTurn().postprocess();
            setDisplayedTurn(0);                // FIXME: initialize object types
        }
        catch (std::exception& e) {
            std::cout << "initialisation error: " << e.what() << "\n";
        }
    } else {
        initGameGlobals(".", Platform::getDataDirectory());
        performInitialisation(OnEnterNoGame, testexpr_modules);
    }
#endif

    while (getline(std::cin, line)) {
        IntTokenizer tok(line);
        IntExprNode* node = 0;
        try {
            IntExpressionParser parser(tok);
            node = parser.parse();
        }
        catch (IntError& e) {
            std::cerr << "parse error: " << e.what() << "\n";
        }
        if (node == 0) {
            // failed
        } else if (tok.getCurrentToken() != tok.tEnd) {
            std::cerr << "parse error; expression not at end\n";
        } else {
            try {
                Ptr<IntBytecodeObject> pbco = new IntBytecodeObject();
                IntBytecodeObject& bco = *pbco;
                node->compileValue(bco, IntCompilationContext().withFlags(IntCompilationContext::LocalContext | IntCompilationContext::AlsoGlobalContext));
                optimize(bco, 1);
                pbco->relocate();
                disassemble(bco);

                // Execute
                IntExecutionContext con("testexpr");
                con.pushNewFrame(new IntExecutionFrame(pbco));
                if (!runTemporaryProcess(con)) {
                    if (con.getState() == IntExecutionContext::psError) {
                        std::cout << "execution error: " << con.getError().what() << "\n";
                    } else {
                        std::cout << "process didn't finish, status is " << con.getState() << "\n";
                    }
                } else {
                    IntValue* result = con.getResult();
                    if (!result)
                        std::cout << "Result is null\n";
                    else
                        std::cout << "Result is " << result->toString(true) << "\n";
                }
            }
            catch (IntError& e) {
                std::cerr << "compile error: " << e.what() << "\n";
            }
        }
        delete node;
    }
}
