/**
  *  \file tests/teststmt.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <iostream>
#include <istream>
#include <string.h>
#include "cpluslib/string.h"
#include "util/console.h"
#include "game/global.h"
#include "game/config.h"
#include "game/spec.h"
#include "game/fcode.h"
#include "game/mission.h"
#include "game/racename.h"
#include "game/coord.h"
#include "client/chartusr.h"
#include "game/stat.h"
#include "game/team.h"
#include "game/load.h"
#include "int/intglobal.h"
#include "int/if/initif.h"
#include "int/csource.h"
#include "cpluslib/format.h"
#include "int/statement.h"
#include "int/bytecode.h"
#include "int/exec.h"
#include "int/interror.h"
#include "io/textfile.h"
#include "int/compound.h"
#include "util/modinit.h"
#include "game/reg.h"
#include "int/if/conif.h"
#include "game/gamedirs.h"
#include "int/args.h"
#include "int/optimize.h"
#include "int/process.h"
#include "game/sel.h"
#include "int/vmio.h"
#include "int/if/ifutil.h"
#include "io/streams.h"
#include "io/memstream.h"

static void
disassemble(IntBytecodeObject& bco)
{
    int lastLine = 0;
    for (IntBytecodeObject::pc_t i = 0; i < bco.getNumInstructions(); ++i) {
        int lineNo = bco.getLineNumber(i);
        if (lineNo != 0 && lineNo != lastLine) {
            std::cout << (format("%4d: <%02X.%02X.%04X>     %-50s ; line %d\n", i, bco(i).major, bco(i).minor, bco(i).arg)
                          << bco.getDisassembly(i)
                          << lineNo);
            lastLine = lineNo;
        } else {
            std::cout << format("%4d: <%02X.%02X.%04X>     ", i, bco(i).major, bco(i).minor, bco(i).arg)
                      << bco.getDisassembly(i) << "\n";
        }
    }
}

class TestIndexable : public IntIndexableValue {
 public:
    IntValue* get(IntArgBlock& args)
        {
            std::cout << "get(";
            while (args.getNumArgs() > 0) {
                std::cout << IntValue::toStringOf(args.getNext(), true);
                if (args.getNumArgs())
                    std::cout << ",";
            }
            std::cout << ")\n";
            return 0;
        }
    void set(IntArgBlock& args, IntValue* value)
        {
            std::cout << "set(";
            while (args.getNumArgs() > 0) {
                std::cout << IntValue::toStringOf(args.getNext(), true);
                if (args.getNumArgs())
                    std::cout << ",";
            }
            std::cout << ") = " << IntValue::toStringOf(value, true) << "\n";
        }
    IntContext* makeFirstContext()
        { return 0; }
    IntValue* clone()
        { return new TestIndexable(); }
    string_t toString(bool)
        { return "TESTER"; }
    void store(IntTagNode&, Stream&)
        { }
};

static void
saveSelf(IntExecutionContext& exc, IntArgBlock& args)
{
    string_t name;
    args.checkArgumentCount(1);
    if (!checkStringArg(name, args.getNext()))
        return;

    // We must remove the IntSimpleProcedureValue referring to us, which is not serializable.
    exc.dropValue();
    exc.pushNewValue(0);

    Ptr<Stream> s = Stream::openFile(name, Stream::C_CREATE);
    const char header[] = { 'C', 'C', 'v', 'm',
                            '1', '1', '-', '1', '1', '-', '1', '1', '1', '1', '1', '1', ':', '1', '1', ':', '1', '1',
                            26, 100,
                            4, 0,
                            1, 0,
                            7, 0 };
    CTASSERT(sizeof(header) == 30);

    IntVMSaveContext sc;
    sc.addProcess(exc);
    s->writeT(header, sizeof(header));
    sc.save(*s);
}

static void
doit(IntCommandSource& cs, bool list)
{
    IntStatementCompilationContext scc;

    // Compile
    IntExecutionContext con("teststmt");
    IntBCORef bco = new IntBytecodeObject();
    IntStatementCompiler::StatementResult res;
    if (list)
        res = IntStatementCompiler(cs).compileList(*bco, scc);
    else
        res = IntStatementCompiler(cs).compile(*bco, scc.withExecutionContext(&con));
    // disassemble(*bco); printf("----\n");
    optimize(*bco, 1);
    bco->relocate();

    // Dump BCO
    disassemble(*bco);
    switch (res) {
     case IntStatementCompiler::Terminator:         std::cout << "** Terminator\n"; break;
     case IntStatementCompiler::EndOfInput:         std::cout << "** EOF\n";        break;
     case IntStatementCompiler::CompiledStatement:  std::cout << "** Statement\n";  break;
     case IntStatementCompiler::CompiledBlock:      std::cout << "** Block\n";      break;
     case IntStatementCompiler::CompiledExpression: std::cout << "** Expression\n"; break;
    }

    // Run it
    con.pushNewFrame(new IntExecutionFrame(bco));
    if (runTemporaryProcess(con)) {
        if (res == IntStatementCompiler::CompiledExpression) {
            IntValue* result = con.getResult();
            if (!result)
                std::cout << "Result is null\n";
            else
                std::cout << "Result is " << result->toString(true) << "\n";
        }
    } else {
        if (con.getState() == IntExecutionContext::psError) {
            std::cout << "execution error: " << con.getError().what() << "\n"
                      << "trace: " << con.getError().getTrace() << "\n";
        } else {
            std::cout << "process didn't finish, status is " << con.getState() << "\n";
        }
    }
}

static const ModuleInitializer*const teststmt_modules[] = {
    mod_base_console,
    mod_int,
    mod_intgame_common,
    mod_intgame_console,
    mod_racenames,
    mod_spec,
    mod_team,
    mod_missions,
    mod_reg,
    mod_stat,
    mod_fcode,
    mod_config,
    mod_sel,
    0
};

int main(int argc, char** argv)
{
    (void) argc;
    (void) argv;

    performInitialisation(OnStartup, teststmt_modules);

    string_t line;

#if 1
    if (argv[1]) {
        try {
            initGameGlobals(Directory::getDirName(argv[1]), Platform::getDataDirectory());
            performInitialisation(OnEnterDirectory | OnEnterGame, teststmt_modules);
            initCoordinates();

            Ptr<Stream> pxy = spec_file_dir->openFile("xyplan.dat", Stream::C_READ);
            getCurrentTurn().init();
            getCurrentTurn().loadPlanetXY(*pxy);

            // HACK!
            int pid = atoi(Directory::getFileName(argv[1]).c_str() + 6);

            Ptr<Stream> rst = Stream::openFile(argv[1], Stream::C_READ);
            loadResult(getCurrentTurn(), *rst, pid);

            setPlayerIds(pid, pid);
            getCurrentTurn().postprocess();
            setDisplayedTurn(0);                // FIXME: initialize object types
        }
        catch (std::exception& e) {
            std::cout << "initialisation error: " << e.what() << "\n";
        }
    } else {
        initGameGlobals(".", Platform::getDataDirectory());
        performInitialisation(OnEnterNoGame, teststmt_modules);
    }
#endif

    std::cout << "Special commands:\n"
              << "  !FILE           load file (beware of special compiler flags)\n"
              << "  ?SUB            disassemble subroutine\n"
              << "  .help           more info\n";
    defineGlobalValue("TESTER", new TestIndexable());
    defineGlobalProcedure("SAVEME", saveSelf);
    while (getline(std::cin, line)) {
        try {
            if (line.size() && line[0] == '!') {
                // Special case: '!file' to load a file
                Ptr<Stream> s = Stream::openFileNT(line.substr(1), Stream::C_READ);
                if (!s)
                    throw IntError("cannot open this file");
                TextFile tf(*s);
                IntFileCommandSource fcs(tf);
                doit(fcs, true);
            } else if (line.size() && line[0] == '?') {
                // Special case: '?fun' to disassemble a function
                string_t fname = strUCase(line.substr(1));
                IntVariableNames::index_t index = global_property_names.lookup(fname);
                if (index == global_property_names.nil)
                    throw IntError("unknown");
                IntSubroutineValue* sv = dynamic_cast<IntSubroutineValue*>(global_values[index]);
                if (!sv)
                    throw IntError("not a subroutine");
                disassemble(*sv->getBytecodeObject());
            } else if (line == ".help") {
                std::cout << "  !FILE           load file (beware of special compiler flags)\n"
                          << "  ?SUB            disassemble subroutine\n"
                          << "  .opt SUB        optimize (and then disassemble) subroutine\n"
                          << "  .list           list all global variables\n"
                          << "  .compile FILE[,OBJ]  compile into object file\n"
                          << "  .load FILE      load object or VM file\n"
                          << "  saveme 'FILE'   within a script writes a VM file\n"
                          << "  tester(X)       is a test 'indexable' to play with\n";
            } else if (line == ".list") {
                for (IntVariableNames::index_t i = 0; i < global_property_names.getNumNames(); ++i)
                    std::cout << i << ": " << global_property_names.lookupAddress(i) << " = " << IntValue::toStringOf(global_values[i], true) << "\n";
            } else if (line.substr(0, 5) == ".opt ") {
                string_t fname = strUCase(line.substr(5));
                IntVariableNames::index_t index = global_property_names.lookup(fname);
                if (index == global_property_names.nil)
                    throw IntError("unknown");
                IntSubroutineValue* sv = dynamic_cast<IntSubroutineValue*>(global_values[index]);
                if (!sv)
                    throw IntError("not a subroutine");
                IntBCORef bco = sv->getBytecodeObject();
                optimize(*bco, 1);
                bco->relocate();
                disassemble(*bco);
            } else if (line.substr(0, 9) == ".compile ") {
                line.erase(0, 9);
                string_t objname;
                string_t::size_type n = line.find(',');
                if (n != line.npos) {
                    objname = line.substr(n+1);
                    line.erase(n);
                } else {
                    objname = line + "c";
                }

                Ptr<Stream> s = Stream::openFileNT(line, Stream::C_READ);
                if (!s)
                    throw IntError("cannot open source file");
                TextFile tf(*s);
                IntFileCommandSource fcs(tf);

                IntStatementCompilationContext scc;
                IntBCORef bco = new IntBytecodeObject();
                IntStatementCompiler(fcs).compileList(*bco, scc);
                optimize(*bco, 1);
                bco->relocate();

                // Prepare save
                IntVMSaveContext vsc;
                uint32_t bco_id = vsc.addBCO(bco);
                uint8_t header[] = { 'C', 'C', 'o', 'b', 'j', 26, 100, 0, 4, 0, bco_id, bco_id>>8, bco_id>>16, bco_id>>24 };

                // Save it
                s = Stream::openFileNT(objname, Stream::C_CREATE);
                if (!s)
                    throw IntError("cannot open object file");
                s->writeT((char*)header, sizeof(header));
                vsc.save(*s);
            } else if (line.substr(0, 6) == ".load ") {
                Ptr<Stream> s = Stream::openFile(line.substr(6), Stream::C_READ);
                char header[14];
                s->readT(header, 14);
                if (memcmp(header, "CCobj", 5) == 0) {
                    printf("*** Object file\n");
                    s->seek(14);
                    IntVMLoadContext vlc;
                    vlc.load(*s, false);
                    IntBCORef initialBCO = vlc.getBCO(getUint32(header+10));
                    if (!initialBCO) {
                        printf("*** Error: invalid initial BCO\n");
                    } else {
                        // Run it
                        IntExecutionContext con("teststmt");
                        con.pushNewFrame(new IntExecutionFrame(initialBCO));
                        if (runTemporaryProcess(con)) {
                            std::cout << "Run successful.\n";
                        } else {
                            if (con.getState() == IntExecutionContext::psError) {
                                std::cout << "execution error: " << con.getError().what() << "\n"
                                          << "trace: " << con.getError().getTrace() << "\n";
                            } else {
                                std::cout << "process didn't finish, status is " << con.getState() << "\n";
                            }
                        }
                    }
                } else if (memcmp(header, "CCvm", 4) == 0) {
                    printf("*** VM file\n");
                    s->seek(30);
                    IntVMLoadContext vlc;
                    vlc.load(*s, true);
                    runRunnableProcesses();
                    killTerminatedProcesses();
                } else {
                    printf("*** unrecognized file type\n");
                }
            } else {
                IntMemoryCommandSource slcs(line);
                doit(slcs, false);
            }
        }
        catch (IntError& e) {
            std::cout << "error: " << e.what() << "\n";
        }
        catch (std::exception& e) {
            std::cout << "exception: " << e.what() << "\n";
        }
    }
}

